# location.rb: written by Takashi SUGA 2002-03-25 v0.4(suchowan@box.email.ne.jp)
#
#  L      OOO   CCC    A   TTTTT I  OOO  N   N
#  L     O   O C   C  A A    T   I O   O NN  N
#  L     O   O C     A   A   T   I O   O N N N
#  L     O   O C   C AAAAA   T   I O   O N  NN
#  LLLLL  OOO   CCC  A   A   T   I  OOO  N   N
#
=begin #########################################################
#=Location                                                     #
\===V̈ʒuvZsNX                                   #
#                                                              #
\==X[p[NX:                                             #
$==Super Class:                                                #
#*Outfit+                                                      #
#                                                              #
\==CN[hĂ郂W[:                             #
$==Modules included:                                           #
#*Location::Common                                             #
#*Location::Ephemeris                                          #
=end ###########################################################

class Location < Outfit

=begin #########################################################
\==Location::Common W[ \bh:                       #
$==Location::Common Module Method:                             #
=end ###########################################################
  module Common
    include Math

    # V萔
    DEG    = (PI/180.0)                 # px1x
    CYCLE  = (2 * PI)                   # px360x
    DAY    =   86400.0                  # /b
    JCENT  =   36525.0                  # EXI/
    BCENT  =   36524.2194               # xbZI/
    EPOCH2 = 2415021.5                  # 1900 I 1.5
    EPOCH4 = 2451545.0                  # 2000 I 1.5
    AU     = 1.49597870E8               # VPʋ
    C0     = 299792.458*86400.0 * JCENT # N
    PSEC   = 360.0*60.0*60.0/(2 * PI)   # p[ZN / AU

=begin ***********
|--- Location::Common.asin, cosd, sind, tand, cosc, sinc, tanc
|  Op֐
=end =============
    def asin(x)  atan2(x, sqrt(1-x*x)) end  # radian  arc sin
    def acos(x)  atan2(sqrt(1-x*x), x) end  # radian  arc cos
    def cosd(x)  cos(x * DEG)          end  # xcos
    def sind(x)  sin(x * DEG)          end  # xsin
    def tand(x)  tan(x * DEG)          end  # xtan
    def cosc(x)  cos(x * CYCLE)        end  # ~Pʂcos
    def sinc(x)  sin(x * CYCLE)        end  # ~Pʂsin
    def tanc(x)  tan(x * CYCLE)        end  # ~Pʂtan

    # ۓzvf^O
    Op     = 0          # Z
    PW0    = 1          # ʑ
    PW1    = 2          # g
    PW2    = 3          # `̏ꍇ̓捀
    Amp    = 3          # U
    Sqr    = 4          # sin,cos̏ꍇ̓捀

    # Z^O
    LIN    =  1
    SIN    =  2
    COS    =  3
    SINT   =  4
    COST   =  5
    SINL   =  6
    COSL   =  7
    SINLT  =  8
    COSLT  =  9
    AcS    = 10

=begin ***********
|--- Location::Common.sum(t, v, l, count)
|  Op֐̘a
|
|    t  Ɨϐ
|    v  vZ鋉i[z
|    v  ̎OvZ
|    count  ̍(ftHg͑SvZ)
=end =============
    def sum(t, v, l=0.0, count=@Precision||0)
      s  = 0
      v[0..count-1].each do |w|
        if w[Op]==1   # 
          ds = w[PW0] + t * (w[PW1] + t * w[PW2])
        else          # Op֐
          ds = w[PW0] + t * (w.length==4 ? w[PW1] : (w[PW1] + t * w[Sqr]))
          ds = w[Amp] * case w[Op]
                        when SIN;   sind(ds)
                        when COS;   cosd(ds)
                        when SINT;  sind(ds) * t
                        when COST;  cosd(ds) * t
                        when SINL;  sind(l+ds)
                        when COSL;  cosd(l+ds)
                        when SINLT; sind(l+ds) * t
                        when COSLT; cosd(l+ds) * t
                        end
        end
        s += ds;
      end
      return s
    end

=begin ***********
|--- Location::Common.poly(x, p)
|  
|
|    x  Ɨϐ
|    p  ̌Wi[z
=end =============
    def poly(x,p)
      s = 0
      p.reverse_each { |c| s = s * x + c }
      return s
    end

=begin ***********
#--- Location::Common.deltaT(date)
\  EXʓ烢TvZĕԂ܂B
\
\    date  EXʓ܂Date*IuWFNg
=end =============
    DeltaT=[                                                     142.00,# 1599

  141.00,140.00,139.00,139.00,139.00,139.00,139.00,140.00,140.00,141.00,# 1600-
  141.00,141.00,141.00,140.00,139.00,138.00,137.00,134.00,132.00,128.00,# 1610-
  124.00,119.00,115.00,110.00,106.00,102.00, 98.00, 95.00, 91.00, 88.00,# 1620-
   85.00, 82.00, 79.00, 77.00, 74.00, 72.00, 70.00, 67.00, 65.00, 63.00,# 1630-
   62.00, 60.00, 58.00, 57.00, 55.00, 54.00, 53.00, 51.00, 50.00, 49.00,# 1640-
   48.00, 47.00, 46.00, 45.00, 44.00, 43.00, 42.00, 41.00, 40.00, 38.00,# 1650-
   37.00, 36.00, 35.00, 34.00, 33.00, 32.00, 31.00, 30.00, 28.00, 27.00,# 1660-
   26.00, 25.00, 24.00, 23.00, 22.00, 21.00, 20.00, 19.00, 18.00, 17.00,# 1670-
   16.00, 15.00, 14.00, 14.00, 13.00, 12.00, 12.00, 11.00, 11.00, 10.00,# 1680-
   10.00, 10.00,  9.00,  9.00,  9.00,  9.00,  9.00,  9.00,  9.00,  9.00,# 1690-

    9.00,  9.00,  9.00,  9.00,  9.00,  9.00,  9.00,  9.00, 10.00, 10.00,# 1700-
   10.00, 10.00, 10.00, 10.00, 10.00, 10.00, 10.00, 11.00, 11.00, 11.00,# 1710-
   11.00, 11.00, 11.00, 11.00, 11.00, 11.00, 11.00, 11.00, 11.00, 11.00,# 1720-
   11.00, 11.00, 11.00, 11.00, 12.00, 12.00, 12.00, 12.00, 12.00, 12.00,# 1730-
   12.00, 12.00, 12.00, 12.00, 13.00, 13.00, 13.00, 13.00, 13.00, 13.00,# 1740-
   13.00, 14.00, 14.00, 14.00, 14.00, 14.00, 14.00, 14.00, 15.00, 15.00,# 1750-
   15.00, 15.00, 15.00, 15.00, 15.00, 16.00, 16.00, 16.00, 16.00, 16.00,# 1760-
   16.00, 16.00, 16.00, 16.00, 16.00, 17.00, 17.00, 17.00, 17.00, 17.00,# 1770-
   17.00, 17.00, 17.00, 17.00, 17.00, 17.00, 17.00, 17.00, 17.00, 17.00,# 1780-
   17.00, 17.00, 16.00, 16.00, 16.00, 16.00, 15.00, 15.00, 14.00, 14.00,# 1790-

   13.70, 13.40, 13.10, 12.90, 12.70, 12.60, 12.50, 12.50, 12.50, 12.50,# 1800-
   12.50, 12.50, 12.50, 12.50, 12.50, 12.50, 12.50, 12.40, 12.30, 12.20,# 1810-
   12.00, 11.70, 11.40, 11.10, 10.60, 10.20,  9.60,  9.10,  8.60,  8.00,# 1820-
    7.50,  7.00,  6.60,  6.30,  6.00,  5.80,  5.70,  5.60,  5.60,  5.60,# 1830-
    5.70,  5.80,  5.90,  6.10,  6.20,  6.30,  6.50,  6.60,  6.80,  6.90,# 1840-
    7.10,  7.20,  7.30,  7.40,  7.50,  7.60,  7.70,  7.70,  7.80,  7.80,# 1850-
    7.88,  7.82,  7.54,  6.97,  6.40,  6.02,  5.41,  4.10,  2.92,  1.82,# 1860-
    1.61,  0.10, -1.02, -1.28, -2.69, -3.24, -3.64, -4.54, -4.71, -5.11,# 1870-
   -5.40, -5.42, -5.20, -5.46, -5.46, -5.79, -5.63, -5.64, -5.80, -5.66,# 1880-
   -5.87, -6.01, -6.19, -6.64, -6.44, -6.47, -6.09, -5.76, -4.66, -3.74,# 1890-

   -2.72, -1.54, -0.02,  1.24,  2.64,  3.86,  5.37,  6.14,  7.75,  9.13,# 1900-
   10.46, 11.53, 13.36, 14.65, 16.01, 17.20, 18.24, 19.06, 20.25, 20.95,# 1910-
   21.16, 22.25, 22.41, 23.03, 23.49, 23.62, 23.86, 24.49, 24.34, 24.08,# 1920-
   24.02, 24.00, 23.87, 23.95, 23.86, 23.93, 23.73, 23.92, 23.96, 24.02,# 1930-
   24.33, 24.83, 25.30, 25.70, 26.24, 26.77, 27.28, 27.78, 28.25, 28.71,# 1940-
   29.15, 29.57, 29.97, 30.36, 30.72, 31.07, 31.35, 31.68, 32.18, 32.68,# 1950-
   33.15, 33.59, 30.00, 34.47, 35.03, 35.73, 36.54, 37.43, 38.29, 39.20,# 1960-
   40.18, 41.17, 42.23, 43.37, 44.49, 45.48, 46.46, 47.52, 48.53, 49.59,# 1970-
   50.54, 51.38, 52.17, 52.96, 53.79, 54.34, 54.87, 55.32, 55.82, 56.30,# 1980-
   56.86, 57.57, 58.31, 59.12, 59.99, 60.79, 61.63, 62.30, 62.97, 63.47,# 1990-

   63.83, 64.09, 64.80, 65.50, 66.20, 66.90, 67.60, 68.30, 69.00, 69.70,# 2000-
    ]
    def deltaT(date)
      rjd = (Date === date) ? date.rjd : date
      h = (rjd-EPOCH4)/JCENT
      t = 2000.0+h*100.0
      y = (t-1599.0).to_i
      if t>=2001.0 
        dt=64.09+0.7*(t-2001.0)
      elsif y>0
        n    = t-(y.to_f+1599.0)
        d0   = DeltaT[y+0]-DeltaT[y-1]
        d1   = DeltaT[y+1]-DeltaT[y+0]
        d2   = DeltaT[y+2]-DeltaT[y+1]
        d10  = d1-d0
        d21  = d2-d1
        d210 = d21-d10
        dt   = DeltaT[y]+n*d1+n*(n-1.0)/4.0*(d10+d21)+n*(n-1.0)*(n-0.5)/6.0*d210
      elsif t<948.0
        dt   = 2715.6+573.36*h+46.5*h*h
      else
        dt   = 50.6+67.5*h+22.5*h*h
      end
      return dt / DAY
    end

=begin ***********
|--- Location::Common.C4(t)
|  EPOCH4(2000N)̃EXIԂ܂B
|
|    t  EXʓ
=end =============
    def C4(t) (t+deltaT(t)-EPOCH4) / JCENT end

=begin ***********
#--- Location::Common.deltaE(t)
#--- Location::Common.deltaP(t)
\  ͓(,)Ԃ܂B
\
\    t  EXʓ
=end =============
    # ͓()
    def deltaE(t)
      return sum(C4(t), [
        [COS   ,  125.04      ,   -1934.136    , +0.00256  ],
        [COS   ,  200.93      ,  +72001.539    , +0.00016  ]
      ], 0.0)
    end

    # ͓()
    def deltaP(t)
      return sum(C4(t), [
        [SIN   ,  125.04      ,   -1934.136    , -0.00478  ],
        [SIN   ,  200.93      ,  +72001.539    , +0.00037  ]
      ], 0.0)
    end

=begin ***********
#--- Location::Common.obl(t)
\  XpԂ܂B
\
\    t  EXʓ
=end =============
    def obl(t)
      return (23.43929 - 0.013004*C4(t) + deltaE(t))
    end
  end # of Location::Common module

=begin #########################################################
|==Location NX {:                                       #
|==Location Class Body:                                        #
=end ###########################################################

  require 'when/moonsun'

  include Common
  include Ephemeris

  @@outfit = {
    'Objs' => [],
    'Hash' => {},
    'Keys' => [
      ['([0-9.]+)([EW])_([0-9.]+)([NS])(_([0-9.]+)H)?',
        proc {|m| Location.new('Title' => m[0],
          'Longitude' =>  m[1].to_i * {'E'=>+1,'W'=>-1}[m[2]],
          'Latitude'  =>  m[3].to_i * {'N'=>+1,'S'=>-1}[m[4]],
          'Height'    => (m[6]||0).to_i)
        }
      ],
      ['Tokyo', proc {|m| Location.new('Title' => m[0],
          'Longitude' => +139.744694,  'Latitude' => +35.654444 )
        }
      ]
    ]
  }
  @@outfit['Key0'] = @@outfit['Keys'].dup

=begin #########################################################
\==NX \bh:                                            #
$==Class Method:                                               #
=end ###########################################################
  class << self

=begin ***********
#--- Location.get(key [, n=1])
\  L[Agp Location+IuWFNgIĕԂ܂B
\
\    key  ϑnW(Location+)L[
\    n    n ԖڂɃ}b` Location+IuWFNgI悤Ɏw
=end =============
    def get(key, n=1)
      case key
      when Location
        return key if (n -= 1) == 0
      when String
        @@outfit['Objs'].each do |item|
          item._title.each do |title|
            return item if key =~ /^#{title}$/ && (n -= 1) == 0
          end
        end
        @@outfit['Keys'].each do |item|
          [*item[0]].each do |title|
            return item[1].call($~.dup) if key =~ /^#{title}$/  && (n-=1) == 0
          end
        end
      end
      return nil
    end

  end # of class method

=begin ***********
#--- Location.new([quality={}])
\  Location+IuWFNg𐶐ĕԂ܂B
\
\    quality  `nbV
=end =============
  def initialize(quality={})
    _copy_quality(quality)
    _qualify
  end

=begin #########################################################
\==pubN CX^X \bh:                           #
$==Public Instance Method:                                     #
=end ###########################################################
  public

=begin ***********
#--- Location#longitude
\  ϑňox/x(o𐳁Ao𕉂Ƃ)Ԃ܂B
=end =============
  def longitude() @Longitude end

=begin ***********
#--- Location#latitude
\  ϑn̈ܓx/x(k܂𐳁A܂𕉂Ƃ)Ԃ܂B
=end =============
  def latitude() @Latitude end

=begin ***********
#--- Location#height
\  ϑn̍x/kmԂ܂B
=end =============
  def height() @Height end

=begin ***********
#--- Location#center?
\  ϑnf̒ŜƂ true A\ʂ̂Ƃ false Ԃ܂B
=end =============
  def center?() @Center ? true : false end

=begin ***********
#--- Location#planet
\  ϑn̂fIDԂ܂B
\  AnɌߑłR[hAnȊOɂ鎖͖ӖłB
=end =============
  def planet() @Planet end

=begin ***********
#--- Location#local_sidereal_time(date)
\  ϑn̒nP/Ԃ܂B
\
\    date  EXʓ܂Date*IuWFNg
=end =============
  def local_sidereal_time(date)
    rjd  = (Date === date) ? date.rjd : date
    c    = C4(rjd)
    @Sid[0] + c * (@Sid[1] + c * @Sid[2])               +
    (@Longitude + cosd(obl(rjd)) * deltaP(rjd)) / 15.0  +
    (rjd+0.5).modulo(1)                         * 24.0
  end
  # func :local_sidereal_time

=begin ***********
#--- Location#equation_of_time(date)
\  ώ/Ԃ܂B
\
\    date  EXʓ܂Date*IuWFNg
=end =============
  def equation_of_time(date)
    rjd  = (Date === date) ? date.rjd : date
    pos = _position(rjd, planet)
    pos = pos.rotate_z(-Planet_Spec[planet][1] / 360.0) if Planet_Spec[planet]
    pos = pos.rotate_z(0.5)
    pos = pos.y_to_r(rjd, self)
    c   = C4(rjd)
    e   = pos.p - (@Sid[0] + c * (@Sid[1] + c * @Sid[2])) / 24.0
    return (0.5 - e.modulo(1))
  end

=begin ***********
#--- Location#obserber_distance
\  ϑn̒nS/kmԂ܂B
=end =============
  def obserber_distance
    @Shape[0] * (@Shape[1] + 
                 @Shape[2] * cosd(2 * @Latitude) + 
                 @Shape[3] * cosd(4 * @Latitude))
  end
  func :obserber_distance

=begin ***********
#--- Location#coordinate(date, system, target)
\  V̂̈ʒuԂ܂B
\
\    date    EXʓ܂Date*IuWFNg
\    system  Wn(1:, 2:ԓ, 3:n)
\    target  ΏېID
=end =============
  def coordinate(date, system, target)
    rjd  = (Date === date) ? date.rjd : date
    pos = _position(rjd, target, planet).parallax(rjd, self) if system > 0
    pos = pos.y_to_r(rjd, self)                              if system > 1
    pos = pos.r_to_h(rjd, self)                              if system > 2
    return pos
  end

=begin ***********
#--- Location#earth_radius
\  ϑ҂n̎a/~Ԃ܂B
=end =============
  def earth_radius
    return 0.25 if @Height <= 0
    air_effect = @Air[1] * @Height / 
                (@Air[2] * @Height + obserber_distance)
    asinc( (1.0 + air_effect) * obserber_distance /
                               (obserber_distance + @Height) )
  end
  func :earth_radius

=begin ***********
#--- Location#star_radius(date, target)
\  V̂̎aԂ܂B
\
\    date    EXʓ܂Date*IuWFNg
\    target  ΏېID
=end =============
  def star_radius(date, target, pos=nil)
    # the Earth
    return earth_raduis if target == planet

    # Other Stars
    rjd  = (Date === date) ? date.rjd : date
    pos  ||= _position(rjd, target, planet).parallax(rjd, self)
    radius = Planet_Spec[target] ? Planet_Spec[target][0] : AU
    return asin(radius / (pos.r*AU)) / CYCLE unless target == 'B'

    # Shadow of the Earth
    (asin(Planet_Spec['3'][0] / pi_moon(rjd)) -
     asin(Planet_Spec['0'][0] / (pos.r * AU)) +
     asin(Planet_Spec['3'][0] / (pos.r * AU))) * 1.02 / CYCLE
  end

=begin ***********
#--- Location#eclipse(date, target1, target2)
\  2V̂̐H֘Aʒu([p/~, H1, H2])Ԃ܂B
\
\    date     EXʓ܂Date*IuWFNg
\    target1  Ώې1ID
\    target2  Ώې2ID
=end =============
  def eclipse(date, target1, target2)
    rjd  = (Date === date) ? date.rjd : date
    pos1 = _position(rjd, target1, planet).parallax(rjd, self)
    pos2 = _position(rjd, target2, planet).parallax(rjd, self)
    rad1 = star_radius(rjd, target1, pos1)
    rad2 = star_radius(rjd, target2, pos2)
    cos  = sinc(pos1.t) * sinc(pos2.t) + 
           cosc(pos1.t) * cosc(pos1.t) * cosc(pos1.p-pos2.p)
    d    = acos(cos) / CYCLE
    e    = rad1 + rad2 - d
    return d, e / (2 * rad1), e / (2 * rad2)
  end

=begin ***********
#--- Location#moon_visibility(date)
\  vɌ邩ǂԂ܂(:, :Ȃ)B
\  ɋ߂ꍇ͍lĂ܂BɐV̏̊mFɗp܂B
\
\    date    nߒʓ܂Date*IuWFNg
=end =============
  def moon_visibility(date)
    t = day_event(date, +1, target='0', height='Z')
    h = coordinate(t, 3, 'A').h * 360.0
    p = eclipse(t, '0', 'A')[0] * 360.0
    n = 4.0 - 0.1*p
    f = p < 40.0 ? 4.0 + n*(n+1)/2 : 4.0
    return (h / f - 1.0)
  end

=begin ***********
#--- Location#day_event(date, event [, target='0' [, height=-air_effect]])
\  V̂ov쒆Ԃ܂B
\  date l̏ꍇ̓EXʓœԂ܂B
\  dateDate*IuWFNg̏ꍇDate*IuWFNgœԂ܂B
\
\    date    nߒʓ܂Date*IuWFNg
\    event   o: -1, 쒆: 0, v: +1
\    target  ΏېID(ftHg͑z)
\    height  臒lx/x(w '0':z̏[n, 'T':閾/)
=end =============
  def day_event(date, event, target='0', height=nil)
    sdn = (Date === date) ? date.sdn : date
    t0  = sdn - longitude / 360.0
    dt  = coordinate(t0, 2, target).p - local_sidereal_time(t0) / 24.0
    dt += 0.25 * event
    dt -= (dt+0.5).floor
    if event == 0
      h = nil
    else
      height = target unless height
      if String === height
        h = - @Zeros[height.upcase]/360.0
        rase ArgumentError, 'invalid height string' unless h
      else
        h = height / 360.0
      end
    end

    h -= 0.25 - earth_radius + @Zeros['A']/360.0
    h += (1.0 - @Height / @Air[0]) * @Zeros['A']/360.0 if @Height < @Air[0]

    t = ::Ephemeris.root(h, t0 + dt) {|t|
      coordinate(t, 3, target).h
    }
    return t unless Date === date
    return Date::Extended.new_date(date.cal, t, date.timezone)
  end

=begin ***********
#--- Location#sun_rise(date [, height=-air_effect])
#--- Location#sun_set(date  [, height=-air_effect])
#--- Location#sun_noon(date)
\  zov쒆Ԃ܂B
\  date l̏ꍇ̓EXʓœԂ܂B
\  dateDate*IuWFNg̏ꍇDate*IuWFNgœԂ܂B
\
\    date    nߒʓ܂Date*IuWFNg
\    height  臒lx/x(w '0':z̏[n, 'T':閾/)
=end =============
  def sun_rise(date, height=nil) day_event(date, -1, target='0', height) end
  def sun_set(date, height=nil)  day_event(date, +1, target='0', height) end
  def sun_noon(date)             day_event(date,  0, target='0')         end

=begin ***********
#--- Location#year_event(date, event, target [,hs=air_effect [,bs=given by table])
\  P̏ovƑz̈ʒu֌WɊւCxg̓Ԃ܂B
\  date l̏ꍇ̓EXʓœԂ܂B
\  dateDate*IuWFNg̏ꍇDate*IuWFNgœԂ܂B
\
\    date    EXʓ܂Date*IuWFNg
\    event   first visible rising : 0 (=wAJECWO)
\            last  visible rising : 1
\            last  visible setting: 2
\            first visible setting: 3
\    target  ΏېID(-FK5\ԍ)
\    hs      P̍x/x(nオ)
\    bs      z̕p/x(nオ, Ēʏ͐ł)
=end =============
  Sgn = [[-1,+1],[-1,-1],[+1,-1],[+1,+1]]
  Bs  = [[11.0,  7.0, 11.0,  7.0],
         [11.0,  7.0, 11.0,  7.0],
         [11.0,  7.0, 11.0,  7.0],
         [14.0,  8.5, 14.0,  8.5],
         [16.0, 10.0, 16.0, 10.0],
         [17.0, 14.0, 17.0, 14.0],
         [17.0, 17.0, 17.0, 17.0]]
  def year_event(date, event, target, hs=-@Zeros['A'], bs=nil)
    rjd = (Date === date) ? date.rjd : date
    cn  = lambda_sun(rjd).floor

    # difference of ecliptic longitude bwtween zenith and target star 
    # when the event happens
    pos = coordinate(rjd, 2, target)
    dp  = (sind(hs) - sind(latitude) * sinc(pos.t)) / 
                     (cosd(latitude) * cosc(pos.t))
    return nil if dp.abs >= 1
    dp  = acos(dp) / CYCLE

    # The Sun's height when the event happens
    unless bs
      mag = Star_Spec[-target.to_i]['Mag']
      bs  = Bs[mag<-0.5 ?  0 : (mag<4.5 ? (mag+0.5).floor : 6)][event]
    end

    zenith = Coordinate.polar(1, latitude/360.0, pos.p+Sgn[event][0]*dp).
             r_to_y(rjd, self)
    s   = sind(bs)/cosc(zenith.t)
    arc = asin(s) / CYCLE + 0.25
    t   = ::Ephemeris.root(cn+(zenith.p+Sgn[event][1]*arc).modulo(1),rjd) {|t|
            lambda_sun(t)
         }
    t   = day_event((t+0.5+longitude/360.0).floor,Sgn[event][0],'0',bs)

    return t unless Date === date
    return Date::Extended.new_date(date.cal, t, date.timezone)
  end

=begin #########################################################
\==uCx[g CX^X \bh:                         #
$==Private Instance Method:                                    #
=end ###########################################################
  private

  Star_Spec   = {}
  Planet_Spec = {
        '0' => [696000.0,    0.00000   ],  #  0: z
        '3' => [  6378.14,   0.0056932 ],  #  3: n
        'A' => [  1738.1,    0.00020   ],  # 10:  
        'B' => [  6378.14,   0.0056932 ]   # 11: {e
  }

=begin ***********
#--- Location#_position(date, target[, base='0'])
#  w̓V̂̈ʒu(zS܂baseV̒S)Ԃ܂B
#
#    date    EXʓ܂Date*IuWFNg
#    target  vZΏۓV('0':z,'3':n,'A':,'B':n̖{e,
#                         'P':k,'V':t_,'Z':V)
#    base    ʒuV('0':z,'3':n)
=end =============
  def _position(date, target, base='0')
    t = (Date === date) ? date.rjd : date

    # SW
    case target
    when '0'  # z
      pos = Coordinate.new(0,0,0,0)
    when '3'  # n
      r, p, c = pi_sun(t), lambda_sun(t), mean_sun(t)
      dp   = 0.5 + Planet_Spec[target][1] / 360 / r
      pos = Coordinate.polar(r, 0, p + dp, c + dp)
    when 'A'  # 
      pos = Coordinate.polar(pi_moon(t)/AU, beta_moon(t),
                            lambda_moon(t), mean_moon(t)) + _position(t,'3')
    when 'B'  # {e
      r, p, c = pi_sun(t), lambda_sun(t)+0.5, mean_sun(t)+0.5
      pos = Coordinate.polar(r + pi_moon / AU, 0, p, c)
    when 'V'  # t_
      pos = Coordinate.polar(10000.0*PSEC, 0, 0)
    when 'P'  # k
      pos = Coordinate.polar(10000.0*PSEC, 0.25, 0).r_to_y(t)
    when 'Z'  # V
      pos = Coordinate.polar(10000.0*PSEC, @Latitude / 360.0,
                              local_sidereal_time(t) /  24.0).r_to_y(t)
    when /^-(\d+)/ # P
      no = $1.to_i
      unless Star_Spec[no]
        File.open(__FILE__.sub(/location.rb$/,'nipe.fk5')) do |f|
          while (f.gets)
            a = $_.split(/ +/)
            if a[1].to_i == no
              Star_Spec[no] = {'T0' =>2000.001279,    'Mag'=>a[-3],
                               'PP' =>a[6].to_f,      'dRR'=>0,
                               'TB' =>a[4].to_f / 360,'LL' =>a[2].to_f / 24,
                               'dTB'=>a[5].to_f      ,'dLL'=>a[3].to_f * 15}
              Star_Spec[no]['Mag'].sub!(/^(-?\d\.\d+).*?$/, '\1')
              Star_Spec[no]['Mag']  = Star_Spec[no]['Mag'].to_f
              Star_Spec[no]['PP' ]  = 0.0001 if Star_Spec[no]['PP' ]==0
            end
          end
        end
        raise ArgumentError, 'invalid FK5 number' unless Star_Spec[no]
      end
      
      t0 = (Star_Spec[no]['T0']-1900.0)/100.0;
      t1 = (t-(EPOCH2-0.68648354))/BCENT - t0;
      d  = PSEC / Star_Spec[no]['PP']
      pos = Coordinate.polar(
          d - t1 * Star_Spec[no]['dRR'] / (AU/(BCENT*86400.0)),
          Star_Spec[no]['TB'] + t1 * Star_Spec[no]['dTB'] / (360*36.0),
          Star_Spec[no]['LL'] + t1 * Star_Spec[no]['dLL'] / (360*36.0)).
        precession(t1, t0).
        rotate_x(-obl(t) / 360).
        nutation(t)
    end
    return pos if base == '0'

    # unvSW - st
    tpos = pos.dup
    bpos = _position(t, base)
    pos  = tpos  - bpos
    dp   = pos.p - bpos.p
    if target == '0' && base == '3'
      pos = pos.rotate_z(-Planet_Spec['3'][1] / 360) 
    elsif !(target == 'A' && base == '3')
      tdp = Planet_Spec[target] ? Planet_Spec[target][1] : 0
      bdp = Planet_Spec[base]   ? Planet_Spec[base][1]   : 0
      pos = Coordinate.polar(
        pos.r,
        pos.t - bdp / 360 * sinc(pos.t) * sinc(dp) / bpos.r,
        pos.p + bdp / 360 / cosc(pos.t) * cosc(dp) / bpos.r -
                tdp / 360 * cosc(pos.p - tpos.p)   / tpos.r,
        pos.c)
    end
    return pos
  end

=begin ***********
#--- Location#_qualify
\  Location+IuWFNg̐`ݒ肵܂B
\
\  @Title     = Location.get ̃L[ƂȂV̈ʒu
\  @Longitude = ϑňox/x(o𐳁Ao𕉂Ƃ)
\  @Latitude  = ϑn̈ܓx/x(k܂𐳁A܂𕉂Ƃ)
\  @Height    = ϑn̍x/km
\  @Center    = ϑnV̂̒SƂƂ^Ƃ
\  @Planet    = ϑn̂V̂ID(nbVPlanet_Spec ̃L[)
\  @Axis      = ϑn̂V̂̒n̖k(2000N)`nbV
\  @Shape     = ϑn̂V̂̑傫ƌ``z
\  @Air       = ϑn̂V̂̑Cʂ`z
\  @Sid       = ϑn̒nP/(2000N)߂z
\  @Zeros     = V̂̏ovvZɗpéu[vx`nbV
=end =============
  def _qualify
    @Height ||=  0
    @Planet ||= '3'
    @Shape  ||= [6378.14, 0.998327112, +0.001676399, -0.000003519, -11.514/60]
    @Air    ||= [10.0, 0.40, 20.0]
    @Sid    ||= [6.697375, 2400.0513369, 0.0000259]
    @Zeros  ||= {'Z'=>0, 'A'=>0.58555, '0'=>0.85255, 'T'=>7.36111}

    if @Title
      @Title = [*@Title]
      @@outfit['Objs'].push(self)
      if @Title[0] =~ /_(SS|SB)$/
        @Bija = $1 == 'SB'
        class << self; include India::Ephemeris end
      end
    end
  # super
  end

  ##############################################################
  # V̂̈ʒuW
  class Coordinate

    include Common

    class << self

      # ɍW
      def polar(r, t, p, c=nil)
        c = Math.cos(2*Math::PI*t)
        s = Math.sin(2*Math::PI*t)
        new(r * c * Math.cos(2*Math::PI*p), r * c * Math.sin(2*Math::PI*p), 
            r * s, c||p)
      end

    end # of class method

    # 𒼐W
    def initialize(x, y, z, c=nil)
      @x, @y, @z, @c = x, y, z, c
      @c = p unless c
    end

    attr_reader :x, :y, :z, :c

    def r() sqrt(@x * @x + @y * @y + @z * @z) end
    def t() return atan2(@z, sqrt(@x * @x + @y * @y))/CYCLE end
    def p() return atan2(@y ,@x)/CYCLE end
    def a() 0.5-p end
    def cn()
      rem, quot = (@c+0.5).moddiv(1)
      return quot-1 if @p > rem - 0.25
      return quot+1 if @p < rem - 0.75
      return quot
    end
    def n() Cycle.new(cn, p) end
    func :r, :t, :p, :a, :cn, :n
    alias_method :d, :r
    alias_method :h, :t

    # vfANZX
    def [](element)
      eval(element.downcase)
    end

    # @
    def + (other)
      unless Coordinate === other
        raise TypeError, 'operand should be Coordinate'
      end
      type.new(@x + other.x, @y + other.y, @z + other.z, @c)
    end

    # @
    def - (other)
      unless Coordinate === other
        raise TypeError, 'operand should be Coordinate'
      end
      type.new(@x - other.x, @y - other.y, @z - other.z, @c)
    end

    # X Ƃ]
    def rotate_x(t)
      c = cosc(t)
      s = sinc(t)
      type.new(@x,  @y * c - @z * s,  @y * s + @z * c, @c)
    end

    # Y Ƃ]
    def rotate_y(t)
      c = cosc(t)
      s = sinc(t)
      type.new(@z * s + @x * c,  @y,  @z * c - @x * s, @c)
    end

    # Z Ƃ]
    def rotate_z(t)
      c = cosc(t)
      s = sinc(t)
      type.new(@x * c - @y * s,  @x * s + @y * c,  @z, @c + t)
    end

    # ͓
    def nutation(t)
      rotate_z(deltaP(t)/360.0).rotate_x(deltaE(t)/360.0)
    end

    # ΍
    def precession(t1, t0)
      return self if t.abs >= 0.25

      b1 = [0.0, +0.302/3600.0, +0.018/3600.0]
      b2 = [0.0, +0.791/3600.0, +0.001/3600.0]
      b3 = [0.0, -0.462/3600.0, -0.042/3600.0]

      b1[0] = (2304.250/3600.0) + (1.396/3600.0) * t0
      b2[0] = poly(t1, b1)
      b3[0] = (2004.682/3600.0) - (0.853/3600.0) * t0

      z0  = t1 * b2[0] / 360.0
      zt  = t1 * poly(t1, b2) / 360.0
      th  = t1 * poly(t1, b3) / 360.0
      aA  = p + z0
      aB  = th / 2
      cA  = cosc(aA)
      sA  = sinc(aA)
      tB  = tanc(aB)
      q   = sinc(th)*(tanc(t) + tB*cA)
      dRA = atan2(q*sA, 1-q*cA) / CYCLE
      dDC = atan2(tB*(cA-sA*tanc(dRA/2)), 1) / CYCLE

      type.polar(r,  t + 2 * dDC,  p + dRA + z0 + zt)
    end

    # nSʒu -> Sʒu(ϑnSʒu)
    def parallax(t, loc)
      return self if loc.center?

      self - Coordinate.polar((loc.obserber_distance + loc.height) / AU,
            (loc.latitude + loc['Shape'][4] * sind(2 * loc.latitude)) / 360.0,
            +loc.local_sidereal_time(t) / 24.0 ).r_to_y(t, loc)
    end

    # ԓW̍΍
    def axis(t, loc=nil)
      return nil unless loc
      x  = loc['Axis']
      return nil unless x
      t0 = (x[0]-1900.0)/100.0
      t1 = (t-(EPOCH2-0.68648354))/BCENT - t0
      type.polar((x[1][0] + t1 * x[2][0] / 360),
                 (x[1][1] + t1 * x[2][1] / 360),
                 (x[1][2] + t1 * x[2][2] / 360)).
      precession(t1, t0)
    end

    # W -> ԓW
    def y_to_r(t, loc=nil)
      c = rotate_x(+obl(t) / 360)
      if (n = axis(t, loc))
        c = c.rotate_z(-n.p).
            rotate_y(-0.25 + n.t).
            rotate_z(-0.25 + n.r)
      end
      return c
    end

    # ԓW -> W
    def r_to_y(t, loc=nil)
      if (n = axis(t, loc))
        c = rotate_z(+0.25 - n.r).
            rotate_y(+0.25 - n.t).
            rotate_z(+n.p)
      else
        c = self
      end
      return c.rotate_x(-obl(t) / 360)
    end

    # ԓW -> nW
    def r_to_h(t, loc)
      rotate_z(-loc.local_sidereal_time(t) / 24).
      rotate_y((loc.latitude - 90) / 360)
    end
  end

  ##############################################################
  # 
  class Cycle

    include Comparable

    # 
    def initialize(cn, cf) @cn, @cf = cn, cf end
    attr_reader :cn, :cf

    # T
    def c() cn + cf end

    # @
    def - (other)
      unless Cycle === other
        raise TypeError, 'operand should be Cycle'
      end
      (cn - other.cn) + (cf - other.cf)
    end

    # r
    def <=> (other)
      unless Cycle === other
        raise TypeError, 'operand should be Cycle'
      end
      return (cn <=> other.cn) unless cn == other.cn
      return (cf <=> other.cf)
    end
  end
end
