# moonsun.rb: written by Takashi SUGA 2002-03-20 v0.3(suchowan@box.email.ne.jp)
#
#  M     M  OOO   OOO  N   N  SSSS U   U N   N
#  MM   MM O   O O   O NN  N S     U   U NN  N
#  M M M M O   O O   O N N N  SSS  U   U N N N
#  M  M  M O   O O   O N  NN     S U   U N  NN
#  M     M  OOO   OOO  N   N SSSS   UUU  N   N
#
=begin #########################################################
#=MoonSun                                                      #
\===Ƒz̍W                                             #
#                                                              #
=end ###########################################################

class Location < Outfit
  module Ephemeris
    include Common

=begin ***********
|--- Common.Y3(t)
|  EPOCH3 ̃EXNԂ܂B
|
|    t  EXʓ
=end =============
    EPOCH3 = 2442412.5   # 1975 I 0.0
    JYEAR  =     365.25  # EXN/
    def Y3(t) (t+deltaT(t)-EPOCH3) / JYEAR end

=begin ***********
|--- Calculation.mean_sun, lambda_sun, pi_sun(t)
|  z̍W(ωoAoA/AU)Ԃ܂B
|
|    t  EXʓ
=end =============
    # z̕ωo
    def mean_sun(t)
      t = Y3(t)
      return sum(t,P3L,0.0,1) / 360.0 + 1974.0
    end

  # z̎o
    S0=[356.531, 359.990504, -1.49416E-8] # z̒S̈ʑ
    S1=[  1.9159, -4.8E-5  , -1.44444E-9] # z̒S̐U
    def lambda_sun(t)
      t = Y3(t)
      return (sind(S0[0]+t*(S0[1]+t*S0[2]))*(S1[0]+t*(S1[1]+t*S1[2])) +
              sum(t,P3L)) / 360.0 + 1974.0
    end

    # z̋^VP
    def pi_sun(t)
      return 10.0**(0.000030+sum(Y3(t),P3Q))
    end

=begin ***********
|--- Calculation.mean_moon, lambda_moon, beta_moon, pi_moon(t)
|  ̍W(ωoAoA܁A/km)Ԃ܂B
|
|    t  EXʓ
=end =============
    # ̕ωo
    def mean_moon(t)
      t = Y3(t)
      return sum(t,P0L,0.0,1) / 360.0 + 38770.0
    end

    # ̎o
    def lambda_moon(t)
      t = Y3(t)
      return sum(t,P0L,sind(sum(t,P0dL)))  / 360.0 + 38770.0
    end

    # ̎
    def beta_moon(t)
      t = Y3(t)
      return sum(t,P0B,sum(t,P0dB)) / 360.0
    end

    #̋
    def pi_moon(t)
      return 6378.14/((sum(Y3(t),P0P)+0.9507)*DEG)
    end

  ##############################################################
  # z̒nSo
  P3L = [
    [LIN   ,  279.0357964 ,     360.007693787, 3.02500E-8],
    [SIN   ,  353.06      ,     719.981008 ,   0.02009 ,-2.98832E-8],
    [SIN   ,  248.64      ,     -19.34136  ,  -0.0048  ],
    [SIN   ,  285.0       ,     329.6446699,   0.0020  ,+8.0E-9    ],
    [SIN   ,  334.2       ,   -4452.671109 ,   0.0018  ,-0.1634E-6 ],

    [SIN   ,  293.7       ,      -0.2020   ,   0.00178 ],
    [SIN   ,  242.4       ,     450.3688   ,   0.0015  ,+1.4E-9],
    [SIN   ,  211.1       ,     225.1844   ,   0.0013  ,+0.7E-9],
    [SIN   ,  208.0       ,     659.2893   ,   0.0008  ,+1.6E-8],
    [SIN   ,   53.5       ,      90.3783   ,   0.0007  ],

    [SIN   ,   12.1       ,     -30.35     ,   0.0007  ],
    [SIN   ,  239.1       ,     337.18     ,   0.0006  ],
    [SIN   ,   10.1       ,      -1.50     ,   0.0005  ],
    [SIN   ,   99.1       ,     -22.81     ,   0.0005  ],
    [SIN   ,  264.8       ,     315.56     ,   0.0004  ],

    [SIN   ,  233.8       ,     299.30     ,   0.0004  ],
    [SIN   ,  198.1       ,     720.02     ,  -0.0004  ],
    [SIN   ,  349.6       ,    1079.97     ,   0.0003  ],
    [SIN   ,  241.2       ,     -44.43     ,   0.0003  ]
  ]

  # z̓ȃΐ
  P3Q = [
    [COS   ,  356.53      ,     359.990504 ,  -0.007261 ,-1.49416E-8],
    [COST  ,  356.53      ,     359.990504 ,   0.0000002,-1.49416E-8],
    [COS   ,  353.1       ,     719.981008 ,  -0.000091 ,-2.98832E-8],
    [COS   ,  205.8       ,    4452.671109 ,   0.000013 ,-0.1634E-6 ],
    [COS   ,   62.0       ,     450.4      ,   0.000007 ,+1.4E-9],
    [COS   ,  105.0       ,     329.6      ,   0.000007 ,+8.0E-9]
  ]

  ##############################################################
  # ̒nSo
  P0dL = [
    [SIN   ,   93.8       ,      -1.33     ,  0.0040   ],
    [SIN   ,  248.6       ,     -19.34     ,  0.0020   ],
    [SIN   ,   66.0       ,       0.2      ,  0.0006   ],
    [SIN   ,  249.0       ,     -19.3      ,  0.0006   ],
  ]

  P0L = [
    [LIN   ,  124.8757417 ,  4812.6788201  , -1.330000E-7],
    [SINL  ,  338.9158263 ,  4771.9886313  ,  6.2887   ,+8.99400E-7],
    [SIN   ,  107.236832  ,   -4133.3536154,  1.2740   ,+1.22612E-6],
    [SIN   ,   51.678963  ,    8905.3422467,  0.6583   ,-0.32672E-6],
    [SIN   ,  317.831653  ,    9543.9772627,  0.2136   ,+1.7988E-6 ],
    [SIN   ,  176.528826  ,     359.990504 ,  0.1856   ,-1.561E-8  ],

    [SIN   ,  292.463     ,    9664.0403505,  0.1143   ,-0.6812E-6 ],
    [SIN   ,   86.16      ,     638.63475182, 0.0588   ,+2.1256E-6 ],
    [SIN   ,  103.78      ,   -3773.36305278, 0.0572   ,-1.22769E-6],
    [SIN   ,   30.58      ,   13677.331    ,  0.0533   ],
    [SIN   ,  124.86      ,   -8545.352    ,  0.0459   ],

    [SIN   ,  342.38      ,    4411.998    ,  0.0410   ],
    [SIN   ,   25.83      ,    4452.671    ,  0.0348   ],
    [SIN   ,  155.45      ,    5131.979    ,  0.0305   ],
    [SIN   ,  240.79      ,     758.698    ,  0.0153   ],
    [SIN   ,  271.38      ,   14436.029    ,  0.0125   ],

    [SIN   ,  226.45      ,   -4892.052    ,  0.0110   ],
    [SIN   ,   55.58      ,  -13038.696    ,  0.0107   ],
    [SIN   ,  296.75      ,   14315.966    ,  0.0100   ],
    [SIN   ,   34.5       ,   -8266.71     ,  0.0085   ],
    [SIN   ,  290.7       ,   -4493.34     ,  0.0079   ],

    [SIN   ,  228.2       ,    9265.33     ,  0.0068   ],
    [SIN   ,  133.1       ,     319.32     ,  0.0052   ],
    [SIN   ,  202.4       ,    4812.66     ,  0.0050   ],
    [SIN   ,   68.6       ,     -19.34     ,  0.0048   ],
    [SIN   ,   34.1       ,   13317.34     ,  0.0040   ],

    [SIN   ,    9.5       ,   18449.32     ,  0.0040   ],
    [SIN   ,   93.8       ,      -1.33     ,  0.0040   ],
    [SIN   ,  103.3       ,   17810.68     ,  0.0039   ],
    [SIN   ,   65.1       ,    5410.62     ,  0.0037   ],
    [SIN   ,  321.3       ,    9183.99     ,  0.0027   ],

    [SIN   ,  174.8       ,  -13797.39     ,  0.0026   ],
    [SIN   ,   82.7       ,     998.63     ,  0.0024   ],
    [SIN   ,    4.7       ,    9224.66     ,  0.0024   ],
    [SIN   ,  121.4       ,   -8185.36     ,  0.0022   ],
    [SIN   ,  134.4       ,    9903.97     ,  0.0021   ],

    [SIN   ,  173.1       ,     719.98     ,  0.0021   ],
    [SIN   ,  100.3       ,   -3413.37     ,  0.0021   ],
    [SIN   ,  248.6       ,     -19.34     ,  0.0020   ],
    [SIN   ,   98.1       ,    4013.29     ,  0.0018   ],
    [SIN   ,  344.1       ,   18569.38     ,  0.0016   ],

    [SIN   ,   52.1       ,  -12678.71     ,  0.0012   ],
    [SIN   ,  250.3       ,   19208.02     ,  0.0011   ],
    [SIN   ,   81.0       ,   -8586.0      ,  0.0009   ],
    [SIN   ,  207.0       ,   14037.3      ,  0.0008   ],
    [SIN   ,   31.0       ,   -7906.7      ,  0.0008   ],

    [SIN   ,  346.0       ,    4052.0      ,  0.0007   ],
    [SIN   ,  294.0       ,   -4853.3      ,  0.0007   ],
    [SIN   ,   90.0       ,     278.6      ,  0.0007   ],
    [SIN   ,  237.0       ,    1118.7      ,  0.0006   ],
    [SIN   ,   82.0       ,   22582.7      ,  0.0005   ],

    [SIN   ,  276.0       ,   19088.0      ,  0.0005   ],
    [SIN   ,   73.0       ,  -17450.7      ,  0.0005   ],
    [SIN   ,  112.0       ,    5091.3      ,  0.0005   ],
    [SIN   ,  116.0       ,    -398.7      ,  0.0004   ],
    [SIN   ,   25.0       ,    -120.1      ,  0.0004   ],

    [SIN   ,  181.0       ,    9584.7      ,  0.0004   ],
    [SIN   ,   18.0       ,     720.0      ,  0.0004   ],
    [SIN   ,   60.0       ,   -3814.0      ,  0.0003   ],
    [SIN   ,   13.0       ,   -3494.7      ,  0.0003   ],
    [SIN   ,   13.0       ,   18089.3      ,  0.0003   ],

    [SIN   ,  152.0       ,    5492.0      ,  0.0003   ],
    [SIN   ,  317.0       ,     -40.7      ,  0.0003   ],
    [SIN   ,  348.0       ,   23221.3      ,  0.0003   ],
  ]

  # ̒nS
  P0dB = [
    [SIN   ,   68.64      ,     -19.341    ,  0.0267   ],
    [SIN   ,  342.0       ,     -19.36     ,  0.0043   ],
    [SIN   ,   93.8       ,      -1.33     ,  0.0040   ],
    [SIN   ,  248.6       ,     -19.34     ,  0.0020   ],
    [SIN   ,  358.0       ,     -19.4      ,  0.0005   ],
  ]

  P0B = [
    [SINL  ,  236.231     ,    4832.0201248,  5.1281218,-0.3406E-6],
    [SIN   ,  215.147     ,    9604.0088   ,  0.2806   ],
    [SIN   ,   77.316     ,      60.0316   ,  0.2777   ],
    [SIN   ,    4.563     ,   -4073.3220   ,  0.1732   ],
    [SIN   ,  308.98      ,    8965.374    ,  0.0554   ],

    [SIN   ,  343.48      ,     698.667    ,  0.0463   ],
    [SIN   ,  287.90      ,   13737.362    ,  0.0326   ],
    [SIN   ,  194.06      ,   14375.997    ,  0.0172   ],
    [SIN   ,   25.6       ,   -8845.31     ,  0.0093   ],
    [SIN   ,   98.4       ,   -4711.96     ,  0.0088   ],

    [SIN   ,    1.1       ,   -3713.33     ,  0.0082   ],
    [SIN   ,  322.4       ,    5470.66     ,  0.0043   ],
    [SIN   ,  266.8       ,   18509.35     ,  0.0042   ],
    [SIN   ,  188.0       ,   -4433.31     ,  0.0034   ],
    [SIN   ,  312.5       ,    8605.38     ,  0.0025   ],

    [SIN   ,  291.4       ,   13377.37     ,  0.0022   ],
    [SIN   ,  340.0       ,    1058.66     ,  0.0021   ],
    [SIN   ,  218.6       ,    9244.02     ,  0.0019   ],
    [SIN   ,  291.8       ,   -8206.68     ,  0.0018   ],
    [SIN   ,   52.8       ,    5192.01     ,  0.0018   ],

    [SIN   ,  168.7       ,   14496.06     ,  0.0017   ],
    [SIN   ,   73.8       ,     420.02     ,  0.0016   ],
    [SIN   ,  262.1       ,    9284.69     ,  0.0015   ],
    [SIN   ,   31.7       ,    9964.00     ,  0.0015   ],
    [SIN   ,  260.8       ,    -299.96     ,  0.0014   ],

    [SIN   ,  239.7       ,    4472.03     ,  0.0013   ],
    [SIN   ,   30.4       ,     379.35     ,  0.0013   ],
    [SIN   ,  304.9       ,    4812.68     ,  0.0012   ],
    [SIN   ,   12.4       ,   -4851.36     ,  0.0012   ],
    [SIN   ,  173.0       ,   19147.99     ,  0.0011   ],

    [SIN   ,  312.9       ,  -12978.66     ,  0.0010   ],
    [SIN   ,    1.0       ,   17870.7      ,  0.0008   ],
    [SIN   ,  190.0       ,    9724.1      ,  0.0008   ],
    [SIN   ,   22.0       ,   13098.7      ,  0.0007   ],
    [SIN   ,  117.0       ,    5590.7      ,  0.0006   ],

    [SIN   ,   47.0       ,  -13617.3      ,  0.0006   ],
    [SIN   ,   22.0       ,   -8485.3      ,  0.0005   ],
    [SIN   ,  150.0       ,    4193.4      ,  0.0005   ],
    [SIN   ,  119.0       ,   -9483.9      ,  0.0004   ],
    [SIN   ,  246.0       ,   23281.3      ,  0.0004   ],

    [SIN   ,  301.0       ,   10242.6      ,  0.0004   ],
    [SIN   ,  126.0       ,    9325.4      ,  0.0004   ],
    [SIN   ,  104.0       ,   14097.4      ,  0.0004   ],
    [SIN   ,  340.0       ,   22642.7      ,  0.0003   ],
    [SIN   ,  270.0       ,   18149.4      ,  0.0003   ],

    [SIN   ,  358.0       ,   -3353.3      ,  0.0003   ],
    [SIN   ,  148.0       ,   19268.0      ,  0.0003   ]
  ]

  # ̎
  P0P = [
    [COS   ,  338.92      ,    4771.98849108, 0.0518   ,+9.19178E-7],
    [COS   ,  287.2       ,   -4133.35355678, 0.0095   ,+1.2262E-6 ],
    [COS   ,   51.7       ,    8905.34223034, 0.0078   ,-0.3268E-6 ],
    [COS   ,  317.8       ,    9543.97698216, 0.0028   ,+1.7988E-6 ],
    [COS   ,   31.0       ,   13677.3      ,  0.0009   ],

    [COS   ,  305.0       ,   -8545.4      ,  0.0005   ],
    [COS   ,  284.0       ,   -3773.4      ,  0.0004   ],
    [COS   ,  342.0       ,    4412.0      ,  0.0003   ]
  ]

  end
end
