# india.rb: written by Takashi SUGA 2002-03-31 v0.4(suchowan@box.email.ne.jp)
#
#  I N   N DDDD  I   A  
#  I NN  N D   D I  A A 
#  I N N N D   D I A   A
#  I N  NN D   D I AAAAA
#  I N   N DDDD  I A   A
#
=begin #########################################################
#=India                                                        #
\===Chn̗                                         #
=end ###########################################################
module India

  module Ephemeris
    include Location::Common
    # Basic Astronomical Constans for Surya-Siddhanta
    E  =       588_465.5-75.8/360 # julian day of Epoch -3101-02-18T00:00:00
    Y  =     4_320_000            # sidereal Years in yuga
    M  =    57_753_336            # siderial Months in yuga
    D  = 1_582_237_500            # sidereal Days in yuga
    R  =       488_203            # Rotation cycles of lunar apogee in yuga
    N  =      -232_238            # rotation cycles of lunar Node in yuga
    B  =           328            # Bija
    P  =           180            # Precession cycles in yuga
    Ep =     1_903_318            # julian day of Precession Epoch 499-01-01
    Ob =  24.0 / 360.0            # Obliquity of the ecliptic
    Am =  90.0 / 360.0            # Agogee of Moon at the beginning of yuga
    Cm = (31.0 + 50.0/60) / 360   # Circle size of Moon for equation of center
    As = (77.0 + 17.0/60) / 360   # Agogee of Sun
    Cs = (13.0 + 50.0/60) / 360   # Circle size of Sun for equation of center

    def days_in_yuga() (D - Y + (@Bija ? B : 0)).to_f end

    def mean_sun(t)     Y * (t - E) / days_in_yuga -  3101  end

    def mean_moon(t)    M * (t - E) / days_in_yuga - 28086  end

    def lambda_sun(t)
      p = mean_sun(t)
      return (p - asin(Cs * sinc(p-As)) / CYCLE)
    end

    def lambda_moon(t)
      p = mean_moon(t)
      a = Am + R * (t - E) / days_in_yuga
      return (p - asin(Cm * sinc(p-a)) / CYCLE)
    end

    def sun_rise(date, height=nil)
      sdn = (Date === date) ? date.sdn : date
      t   = sdn - @Longitude / 360.0 - 0.25
      p   = mean_sun(t) +  P * (t - Ep) / days_in_yuga
      return (t - asin(tand(@Latitude)*tan(asin(sinc(p)*sinc(Ob)))) / CYCLE)
    end
  end

  [['CentralIndia',       +82.5, +23+11/60.0],
   ['NorthIndia',         +82.5, +29.0      ],
   ['Ujjain',             +75.8, +23.2      ],
   ['Agra',               +78.0, +27.2      ],
   ['Ahmedabad',          +72.6, +23.0      ],
   ['Ajmer',              +74.6, +26.5      ],
   ['Aligarh',            +78.1, +27.9      ],
   ['Amritsar',           +74.9, +31.6      ],
   ['Bangalore',          +77.6, +13.0      ],
   ['Bhuvaneswar',        +85.8, +20.2      ],
   ['Bombay',             +72.8, +19.0      ],
   ['Calcutta',           +88.4, +22.6      ],
   ['Chennai',            +80.3, +13.1      ],
   ['Cochin',             +76.2, +10.0      ],
   ['Colombo',            +79.9,  +6.9      ],
   ['Dacca',              +90.4, +23.7      ],
   ['Delhi',              +77.2, +28.6      ],
   ['Hyderabad',          +78.5, +17.4      ],
   ['Jaipur',             +75.8, +26.9      ],
   ['Kathmandu',          +85.2, +27.7      ],
   ['Koch',               +76.2, +10.0      ],
   ['Lahor',              +74.3, +31.6      ],
   ['Madras',             +80.3, +13.1      ],
   ['Mathura',            +77.7, +27.5      ],
   ['Mumbai',             +72.8, +19.0      ],
   ['Mysore',             +76.6, +12.3      ],
   ['Patna',              +85.1, +25.6      ],
   ['Pune',               +73.9, +18.5      ],
   ['Srinagar',           +74.8, +34.1      ],
   ['Thiruvananthapuram', +77.0,  +8.5      ],
   ['Trivandrum',         +77.0,  +8.5      ],
   ['Varanasi',           +83.0, +25.3      ]].each do |place|
     Location.register(place[0]+'(_.*)?') {|m| Location.new(
       'Title' => m[0], 'Longitude' => place[1],  'Latitude' => place[2])
     }
  end

  def self.instance(key)
    case key
    when /^ILSC(M|SS|SB)(A|P|PX)(V|S|VZ|SZ)(_(.+))?$/
      DMD.new(
        'Title'       => key,
        'Ephemeris'   => '30L' + '_' +
                         ($5 || ($1=='M'?'CentralIndia':'Ujjain'))+'_'+ $1,
        'Timezone'    =>  5.5,
        'Calendar'    => key.sub(/C/,'B')
      )
    when /^ILSB(M|SS|SB)(A|P|PX)(V|S|VZ|SZ)(_(.+))?$/
      ILSB.new(
        'Title'       => key,
        'Ephemeris'   => '30L=>12S' + '_' +
                         ($5 || ($1=='M'?'CentralIndia':'Ujjain'))+'_'+ $1,
        'CycleOffset' => {'M'=>+23.25/30, 'SS'=>0.0, 'SB'=>0.0}[$1],
        'IndianStyle' => {'A'=>1, 'P'=>2, 'PX'=>3}[$2],
        'YearEpoch'   => {'V'=>-58,'VZ'=>-57,'S'=>78, 'SZ'=>79}[$3]
      )
    end
  end
end
